;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "ednc" "20251104.1820"
  "Emacs Desktop Notification Center."
  '((emacs "26.1"))
  :url "https://github.com/sinic/ednc"
  :commit "638d73e7c3fe4ffe86199e33fb85fbaa4cf1ab7f"
  :revdesc "638d73e7c3fe"
  :keywords '("unix")
  :authors '(("Simon Nicolussi" . "sinic@sinic.name"))
  :maintainers '(("Simon Nicolussi" . "sinic@sinic.name")))
